#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################
# シナリオ
# ロボットが棚の前まで移動し，右手で商品を把持
# 把持後に人の前まで移動し，商品を提供後に初期
# 位置に戻る
##############################################

import rospy
import math
import copy
import moveit_commander
import moveit_msgs.msg

from geometry_msgs.msg import Quaternion, Pose, PoseStamped, Vector3
from tf.transformations import quaternion_from_euler

import actionlib
from move_base_msgs.msg import MoveBaseAction, MoveBaseGoal

######################################
# GLOBAL
######################################

waypoints_1 = [
    [(0.0, 0.0, 0.0),(0.0, 0.0, 0.0, 1.0)],
    [(-1.717, 0.794, 0.0),(0.0, 0.0, 1.0, 0.045)],
    [(-2.457, 6.352, 0.0),(0.0, 0.0, 0.709, 0.705)]
]

waypoints_2 = [
    [(-2.522, 7.077, 0.0),(0.0, 0.0, 0.699, 0.715)]
]

waypoints_3 = [
    [(-2.457, 6.352, 0.0),(0.0, 0.0, 0.709, 0.705)],
    [(-0.194, 5.370, 0.0),(0.0, 0.0, -0.714, 0.700)],
    [(-2.966, 2.432, 0.0),(0.0, 0.0, 1.000, -0.015)]
]

waypoints_4 = [
    [(0.0, 0.0, 0.0),(0.0, 0.0, 0.0, 1.0)]
]

def goal_pose(pose):
    goal_pose = MoveBaseGoal()
    goal_pose.target_pose.header.frame_id = 'map'
    goal_pose.target_pose.pose.position.x = pose[0][0]
    goal_pose.target_pose.pose.position.y = pose[0][1]
    goal_pose.target_pose.pose.position.z = pose[0][2]
    goal_pose.target_pose.pose.orientation.x = pose[1][0]
    goal_pose.target_pose.pose.orientation.y = pose[1][1]
    goal_pose.target_pose.pose.orientation.z = pose[1][2]
    goal_pose.target_pose.pose.orientation.w = pose[1][3]
    return goal_pose

SLEEP_TIME = 2.0

def euler_to_quaternion(role, pitch, yaw):
    q = quaternion_from_euler(role, pitch, yaw)
    return Quaternion(q[0], q[1], q[2], q[3])


######################################
# MAIN
######################################


def main():
    
    # init node
    rospy.init_node("final_task_node")
    
    # actionlib client declaration 
    client = actionlib.SimpleActionClient('move_base', MoveBaseAction)
    client.wait_for_server()

    # configuration for moveit
    robot = moveit_commander.RobotCommander()
    scene = moveit_commander.PlanningSceneInterface()
    
    # group name
    rarm = moveit_commander.MoveGroupCommander("rarm")
    larm = moveit_commander.MoveGroupCommander("larm")
    upper = moveit_commander.MoveGroupCommander("upper_body")
    rarm_waist = moveit_commander.MoveGroupCommander("rarm_with_waist")

    ## reset pose
    joint_goal = upper.get_current_joint_values()
    for i in range(0,len(joint_goal)):
        joint_goal[i] = 0
    joint_goal[6] = -3
    joint_goal[16] = -3
    upper.go(joint_goal, wait=True)

    # waypoints_1
    # 棚の前まで移動
    for pose in waypoints_1:
        goal = goal_pose(pose)
        client.send_goal(goal)
        client.wait_for_result()

    rospy.sleep(5)

    ## pose goal with rarm
    # 左腕を出す
    joint_goal = upper.get_current_joint_values()
    for i in range(0,len(joint_goal)):
        joint_goal[i] = 0
    joint_goal[6] = -1.57
    joint_goal[16] = -3
    upper.go(joint_goal, wait=True)

    rospy.sleep(10)

    # waypoints_2
    # 棚にあるものを掴みに行く
    for pose in waypoints_2:
        goal = goal_pose(pose)
        client.send_goal(goal)
        client.wait_for_result()

    rospy.sleep(10)

    # waypoints_3
    # 人の前まで移動
    for pose in waypoints_3:
        goal = goal_pose(pose)
        client.send_goal(goal)
        client.wait_for_result()

    rospy.sleep(15)

    ## reset pose
    joint_goal = upper.get_current_joint_values()
    for i in range(0,len(joint_goal)):
        joint_goal[i] = 0
    joint_goal[6] = -3
    joint_goal[16] = -3
    upper.go(joint_goal, wait=True)

    # waypoints_4
    # 初期位置に移動
    for pose in waypoints_4:
        goal = goal_pose(pose)
        client.send_goal(goal)
        client.wait_for_result()
    
if __name__ == '__main__':
    try:
        main()
    except rospy.ROSInterruptException:
        pass
